#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <asm/types.h>
#include <pthread.h>
#include <errno.h>
#include <glib.h>
#include <gio/gio.h>
#include <libkmod.h>
#include <kysec/kysec_communicate.h>

extern void do_syslog(const char *format, ...);
extern GVariant *process_dbus_method(gchar *method, GVariant *parameters, int *err_code);
extern int handling_dbus_error(GError *error);
extern char* get_full_path(char* fpath);

/*
 * 功能：通过ko文件名获取对应模块的名字
 * ko_file: 可以是单个的ko文件名，也可以是全路径的ko文件
 * 返回值：成功时返回"char *"类型的模块名（需要free释放），失败或文件名格式不对时返回NULL
 */
char *get_kmod_name_from_ko(const char *ko_file)
{
#if 1
	if (ko_file == NULL) {
		fprintf(stderr, "Ko file is Null\n");
		return NULL;
	}

	char *really_name = NULL;
	struct kmod_module *kmod = NULL;
	struct kmod_ctx *ctx = NULL;
	int ret = 0;

	ctx = kmod_new(NULL, NULL);
	if (ctx == NULL) {
		fprintf(stderr, "Failed to kmod new\n");
		return NULL;
	}

	ret = kmod_module_new_from_path(ctx, ko_file, &kmod);
	if (ret < 0) {
		fprintf(stderr, "Failed to new kmod from path: %s\n", ko_file);
		goto out;
	}

	really_name = malloc(1024);
	if (really_name == NULL) {
		fprintf(stderr, "Failed to malloc\n");
		goto out1;
	}

	memset(really_name, 0, 1024);
	ret = tkl_kmod_module_get_module_name(kmod, 0, NULL, really_name);
	if (ret < 0) {
		fprintf(stderr, "Failed to get module really name for '%s'\n", ko_file);
		free(really_name);
		really_name = NULL;
	}

out1:
	kmod_module_unref(kmod);
out:
	kmod_unref(ctx);

	return really_name;
#endif
	return NULL;
}

/*
 * 功能：添加需要保护的模块名到模块保护列表文件中
 * kmod_name: 是模块名
 */
int add_to_kmod_protect(char *ko_file)
{
	if (NULL == ko_file) {
		fprintf(stderr, "ko_file is NULL\n");
		return -1;
	}

	int ret = 0;
	char *ko_name = get_kmod_name_from_ko(ko_file);

	if (ko_name == NULL) {
		fprintf(stderr, "get kmod name failed\n");
		return -2;
	}

	char cmd[2048] = {0};

	snprintf(cmd, sizeof(cmd), "add %s", ko_name);
	ret = kysec_kmod_communicate_with_kernel(cmd);
	if (ret < 0) {
		free(ko_name);
		fprintf(stderr, "set kmod pro failed\n");
		return -3;
	}

	char* absolute_path = get_full_path(ko_file);	
	GVariant *parameters = g_variant_new("(s)", absolute_path);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_add_rmmodpro", parameters, &err_code);
	if (NULL == result)	{
		ret = err_code;
		goto rollback;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));
	g_variant_unref(result);
	if (ret != 0)
		goto rollback;

	free(absolute_path);
	return 0;

rollback:
	memset(cmd, 0x00, sizeof(cmd));
	snprintf(cmd, sizeof(cmd), "del %s", ko_name);
	kysec_kmod_communicate_with_kernel(cmd);
	free(ko_name);
	free(absolute_path);

	return ret;
}

/*
 * 功能：从模块保护列表文件中移除不需要保护的模块
 * kmod_name: 是模块名
 */
int remove_from_kmod_protect(char *ko_file)
{
	if (NULL == ko_file) {
		fprintf(stderr, "ko_file is NULL\n");
		return -1;
	}

	int ret = 0;
	char *ko_name = get_kmod_name_from_ko(ko_file);
	if (ko_name == NULL) {
		fprintf(stderr, "get kmod name failed\n");
		return -2;
	}

	char cmd[2048] = {0};
	snprintf(cmd, sizeof(cmd), "del %s", ko_name);
	ret = kysec_kmod_communicate_with_kernel(cmd);
	if (ret < 0) {
		free(ko_name);
		fprintf(stderr, "set kmod pro failed\n");
		return -3;
	}

	char* absolute_path = get_full_path(ko_file);
	GVariant *parameters = g_variant_new("(s)", absolute_path);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_remove_rmmodpro", parameters, &err_code);
	if (NULL == result)	{
		ret = err_code;
		goto rollback;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));
	g_variant_unref(result);
	if (ret != 0)
		goto rollback;

	free(absolute_path);
	return 0;

rollback:
	memset(cmd, 0x00, sizeof(cmd));
	snprintf(cmd, sizeof(cmd), "add %s", ko_name);
	kysec_kmod_communicate_with_kernel(cmd);
	free(ko_name);
	free(absolute_path);

	return ret;
}
