/* gcc kysec_xml.c -o kysec_xml `pkg-config --libs --cflags libxml-2.0` -lcrypto */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/encoding.h>
#include <libxml/xpath.h>
#include <openssl/evp.h>
#include <openssl/sha.h>

#include <kysec/kysec_xml.h>

/* hash: filepath+uid */
static char *get_path_hash(const char *filepath, uid_t uid)
{
	int i = 0;
	unsigned char result[XML_HASH_SIZE/2] = {0};
	char *value = NULL;
	SHA256_CTX ctx;
	char *path_tmp;
	int len = 0;

	if (!filepath)
		return NULL;

	value = malloc(XML_HASH_SIZE + 1);
	if (NULL == value) {
		printf("failed to malloc!\n");
		return NULL;
	}
	memset(value, 0, XML_HASH_SIZE + 1);
	memset(result, 0x0, sizeof result);

	len = strlen(filepath) + 10;
	path_tmp = malloc(len * sizeof(char));
	if (!path_tmp) {
		free(value);
		return NULL;
	}
	memset(path_tmp, '\0', len * sizeof(char));
	snprintf(path_tmp, len * sizeof(char), "%s+%d", filepath, uid);

	SHA256_Init(&ctx);
	SHA256_Update(&ctx, path_tmp, strlen(path_tmp));
	SHA256_Final(result, &ctx);

	for (i = 0; i < XML_HASH_SIZE/2; i++)
		snprintf(value + i*2, 3, "%02x", result[i]);
	value[XML_HASH_SIZE] = '\0';

	free(path_tmp);
	return value;
}

static int kysec_xml_create(char *version)
{
	xmlDocPtr doc;
	xmlNodePtr kysec_root;
	int retval = 0;

	doc = xmlNewDoc(BAD_CAST(version));
	kysec_root = xmlNewNode(NULL, BAD_CAST("kysec_root"));
	xmlDocSetRootElement(doc, kysec_root);

	/* return: n bytes writed in file, -1 is error */
	retval = xmlSaveFile(KYSEC_XML_PATH, doc);
	xmlFreeDoc(doc);
	return retval;
}

static void kysec_xml_error_handle(void)
{
	xmlErrorPtr err;

	err = xmlGetLastError();
	//printf("%s(%d)\n", err->message, err->code);
	if (err->code == XML_IO_LOAD_ERROR || err->code == XML_ERR_DOCUMENT_EMPTY)
		kysec_xml_create("1.0");
}

/* return kysec_xml list, and the list size is count */
struct kysec_xml_kernel *kysec_xml_read_all_data(int *count)
{
	xmlDocPtr doc;
	xmlNodePtr curNode = NULL;
	xmlNodePtr userNode = NULL;
	int nodeCount = 0;
	struct kysec_xml_kernel *xmlData = NULL;
	xmlErrorPtr err;

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return NULL;
	}

	/* get root */
	curNode = xmlDocGetRootElement(doc);
	if (!curNode) {
		xmlFreeDoc(doc);
		return NULL;
	}

	if (xmlStrcmp(curNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		xmlFreeDoc(doc);
		return NULL;
	}

	/* kysec_root children */
	userNode = curNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		/* check if have user */
		if (!userID)
			goto user_next;

		xmlFree(userID);
		/* find user node */
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlNodePtr childNode;
				xmlAttrPtr attr;
				xmlChar *path = NULL, *hashVal = NULL, *type = NULL;
				int hashLen = 0;

				if (!xmlHasProp(curNode,BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path = xmlGetProp(curNode, BAD_CAST"path");

				/* kysec_node children */
				childNode = curNode->xmlChildrenNode;
				while (childNode != NULL) {
					if (!xmlStrcmp(childNode->name, (const xmlChar *)"type"))
						type = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"hash"))
						hashVal = xmlNodeGetContent(childNode);
					childNode = childNode->next;
				}

				//printf("path='%s', hash='%s', type='%s'\n", path, hashVal, type);
				if (path && hashVal && type) {
					nodeCount++;
					xmlData = realloc(xmlData, nodeCount * sizeof(struct kysec_xml));

					hashLen = xmlStrlen(hashVal);
					if (hashLen > XML_HASH_SIZE)
						hashLen = XML_HASH_SIZE;

					memset(xmlData[nodeCount - 1].path, '\0', PATH_MAX);
					memset(xmlData[nodeCount - 1].hash, '\0', XML_HASH_SIZE + 1);
					memcpy(xmlData[nodeCount - 1].path, path, xmlStrlen(path));
					if (hashVal)
						memcpy(xmlData[nodeCount - 1].hash, hashVal, hashLen);
					if (type)
						xmlData[nodeCount - 1].type = atoi((char *)type);
					else
						xmlData[nodeCount - 1].type = OP_UNKNOWN;
				}
				if (path)
					xmlFree(path);
				if (hashVal)
					xmlFree(hashVal);
				if (type)
					xmlFree(type);
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		userNode = userNode->next;
	}

	*count = nodeCount;
	xmlFreeDoc(doc);

	return xmlData;
}

struct kysec_xml *kysec_xml_read_all_data_with_uid(uid_t uid, int *count)
{
	xmlDocPtr doc;
	xmlNodePtr curNode = NULL;
	xmlNodePtr userNode = NULL;
	xmlNodePtr curUserNode = NULL;
	int nodeCount = 0;
	struct kysec_xml *xmlData = NULL;
	xmlErrorPtr err;

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return NULL;
	}

	/* get root */
	curNode = xmlDocGetRootElement(doc);
	if (!curNode) {
		xmlFreeDoc(doc);
		return NULL;
	}

	if (xmlStrcmp(curNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		xmlFreeDoc(doc);
		return NULL;
	}

	/* kysec_root children */
	userNode = curNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		if (!userID)
			goto user_next;
		if (userID && uid != atoi(userID)) {
			xmlFree(userID);
			goto user_next;
		}

		xmlFree(userID);
		/* find user node */
		curUserNode = userNode;
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlNodePtr childNode;
				xmlAttrPtr attr;
				xmlChar *path = NULL, *hashVal = NULL, *type = NULL;
				xmlChar *name = NULL, *ico = NULL, *server_type = NULL;
				int nameLen = 0, hashLen = 0;

				if (!xmlHasProp(curNode,BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path = xmlGetProp(curNode, BAD_CAST"path");

				/* kysec_node children */
				childNode = curNode->xmlChildrenNode;
				while (childNode != NULL) {
					if (!xmlStrcmp(childNode->name, (const xmlChar *)"type"))
						type = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"hash"))
						hashVal = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"name"))
						name = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"ico"))
						ico = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"server_type"))
						server_type = xmlNodeGetContent(childNode);
					childNode = childNode->next;
				}

				//printf("path='%s', hash='%s', type='%s'\n", path, hashVal, type);
				if (path && hashVal && type) {
					nodeCount++;
					xmlData = realloc(xmlData, nodeCount * sizeof(struct kysec_xml));

					hashLen = xmlStrlen(hashVal);
					nameLen = xmlStrlen(name);
					if (hashLen > XML_HASH_SIZE)
						hashLen = XML_HASH_SIZE;
					if (nameLen > NAME_MAX)
						nameLen = NAME_MAX;

					memset(xmlData[nodeCount - 1].path, '\0', PATH_MAX);
					memset(xmlData[nodeCount - 1].hash, '\0', XML_HASH_SIZE + 1);
					memset(xmlData[nodeCount - 1].name, '\0', NAME_MAX + 1);
					memset(xmlData[nodeCount - 1].ico, '\0', PATH_MAX);
					memcpy(xmlData[nodeCount - 1].path, path, xmlStrlen(path));
					if (hashVal)
						memcpy(xmlData[nodeCount - 1].hash, hashVal, hashLen);
					if (name)
						memcpy(xmlData[nodeCount - 1].name, name, nameLen);
					else
						memcpy(xmlData[nodeCount - 1].name, "unknown", 7);
					if (ico)
						memcpy(xmlData[nodeCount - 1].ico, ico, xmlStrlen(ico));
					else
						memcpy(xmlData[nodeCount - 1].ico, "unknown", 7);
					if (type)
						xmlData[nodeCount - 1].type = atoi((char *)type);
					else
						xmlData[nodeCount - 1].type = OP_UNKNOWN;
					if (server_type)
						xmlData[nodeCount - 1].server_type = atoi((char *)server_type);
					else
						xmlData[nodeCount - 1].server_type = SERV_UNKNOWN;
				}
				if (path)
					xmlFree(path);
				if (hashVal)
					xmlFree(hashVal);
				if (type)
					xmlFree(type);
				if (name)
					xmlFree(name);
				if (ico)
					xmlFree(ico);
				if (server_type)
					xmlFree(server_type);
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		/* have found user node, not need to loop */
		if (curUserNode)
			break;
		userNode = userNode->next;
	}

	*count = nodeCount;
	xmlFreeDoc(doc);

	return xmlData;
}

struct kysec_xml *kysec_xml_read_node_with_uid(uid_t uid, char *filepath)
{
	xmlDocPtr doc;
	xmlNodePtr curNode = NULL;
	xmlNodePtr userNode = NULL;
	xmlNodePtr curUserNode = NULL;
	int nodeCount = 0;
	struct kysec_xml *xmlData = NULL;
	xmlErrorPtr err;

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return NULL;
	}

	/* get root */
	curNode = xmlDocGetRootElement(doc);
	if (!curNode) {
		xmlFreeDoc(doc);
		return NULL;
	}

	if (xmlStrcmp(curNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		xmlFreeDoc(doc);
		return NULL;
	}

	/* kysec_root children */
	userNode = curNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		if (!userID)
			goto user_next;
		if (userID && uid != atoi(userID)) {
			xmlFree(userID);
			goto user_next;
		}

		xmlFree(userID);
		/* find user node */
		curUserNode = userNode;
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlNodePtr childNode;
				xmlAttrPtr attr;
				xmlChar *path = NULL, *hashVal = NULL, *type = NULL;
				xmlChar *name = NULL, *ico = NULL, *server_type = NULL;
				int nameLen = 0, hashLen = 0;

				if (!xmlHasProp(curNode,BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path = xmlGetProp(curNode, BAD_CAST"path");
				if (!path)
					goto cur_next;
				if (xmlStrcmp(path, (xmlChar *)filepath)) {
					xmlFree(path);
					goto cur_next;
				}

				/* found */
				childNode = curNode->xmlChildrenNode;
				while (childNode != NULL) {
					if (!xmlStrcmp(childNode->name, (const xmlChar *)"type"))
						type = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"hash"))
						hashVal = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"name"))
						name = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"ico"))
						ico = xmlNodeGetContent(childNode);
					else if (!xmlStrcmp(childNode->name, (const xmlChar *)"server_type"))
						server_type = xmlNodeGetContent(childNode);
					childNode = childNode->next;
				}

				//printf("path='%s', hash='%s', type='%s'\n", path, hashVal, type);
				if (path && hashVal && type) {
					xmlData = malloc(sizeof(struct kysec_xml));

					hashLen = xmlStrlen(hashVal);
					nameLen = xmlStrlen(name);
					if (hashLen > XML_HASH_SIZE)
						hashLen = XML_HASH_SIZE;
					if (nameLen > NAME_MAX)
						nameLen = NAME_MAX;

					memset(xmlData->path, '\0', PATH_MAX);
					memset(xmlData->hash, '\0', XML_HASH_SIZE + 1);
					memset(xmlData->name, '\0', NAME_MAX + 1);
					memset(xmlData->ico, '\0', PATH_MAX);
					memcpy(xmlData->path, path, xmlStrlen(path));
					if (hashVal)
						memcpy(xmlData->hash, hashVal, hashLen);
					if (name)
						memcpy(xmlData->name, name, nameLen);
					else
						memcpy(xmlData->name, "unknown", 7);
					if (ico)
						memcpy(xmlData->ico, ico, xmlStrlen(ico));
					else
						memcpy(xmlData->ico, "unknown", 7);
					if (type)
						xmlData->type = atoi((char *)type);
					else
						xmlData->type = OP_UNKNOWN;
					if (server_type)
						xmlData->server_type = atoi((char *)server_type);
					else
						xmlData->server_type = SERV_UNKNOWN;
				}
				if (path)
					xmlFree(path);
				if (hashVal)
					xmlFree(hashVal);
				if (type)
					xmlFree(type);
				if (name)
					xmlFree(name);
				if (ico)
					xmlFree(ico);
				if (server_type)
					xmlFree(server_type);
				goto out;
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		/* have found user node, not need to loop */
		if (curUserNode)
			break;
		userNode = userNode->next;
	}
out:
	xmlFreeDoc(doc);

	return xmlData;

}

/* return: 0 success, 1 already exist, -1 error */
int kysec_xml_add_node(uid_t uid, char *path, int type,
		char *name, char *ico, int server_type)
{
	xmlDocPtr doc;
	xmlNodePtr rootNode = NULL;
	xmlNodePtr userNode = NULL;
	xmlNodePtr curUserNode = NULL;
	xmlNodePtr curNode = NULL;
	xmlNodePtr newNode = NULL;
	xmlNodePtr hashNode = NULL;
	xmlNodePtr typeNode = NULL;
	xmlNodePtr nameNode = NULL;
	xmlNodePtr icoNode = NULL;
	xmlNodePtr serverTypeNode = NULL;
	int nodeCount = 0;
	struct kysec_xml *xmlData = NULL;
	int ret = 0;
	char type_tmp[10] = {0};
	char server_type_tmp[10] = {0};
	char *hashVal = NULL;

	if (!path)
		return -1;

	if (type < OP_ALLOW || type > OP_AUTH) {
		printf("The type (%d) is a wrong operation type\n", type);
		return -1;
	}

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return -1;
	}

	/* get root */
	rootNode = xmlDocGetRootElement(doc);
	if (!rootNode) {
		ret = -1;
		goto out;
	}

	if (xmlStrcmp(rootNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		ret = -1;
		goto out;
	}

	/* kysec_root children */
	userNode = rootNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		if (!userID)
			goto user_next;
		if (userID && uid != atoi(userID)) {
			xmlFree(userID);
			goto user_next;
		}

		xmlFree(userID);
		/* find user node */
		curUserNode = userNode;
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlAttrPtr attr;
				xmlChar *path_tmp = NULL;

				if (!xmlHasProp(curNode, BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path_tmp = xmlGetProp(curNode, BAD_CAST"path");
				if (!path_tmp)
					goto cur_next;
				/* already exist */
				if (path_tmp && !xmlStrcmp(path_tmp, BAD_CAST(path))) {
					xmlFree(path_tmp);
					ret = 1;
					goto out;
				}

				xmlFree(path_tmp);
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		/* have found user node, not need to loop */
		if (curUserNode)
			break;
		userNode = userNode->next;
	}

	/* add new node */
	if (!curUserNode) {
		char user_tmp[10] = {0};

		curUserNode = xmlNewNode(NULL, BAD_CAST"kysec_user");
		memset(user_tmp, '\0', sizeof(user_tmp));
		snprintf(user_tmp, sizeof(user_tmp), "%d", uid);
		xmlNewNsProp(curUserNode, NULL, BAD_CAST"userID", BAD_CAST(user_tmp));
		xmlAddChild(rootNode, curUserNode);
	}
	newNode = xmlNewNode(NULL, BAD_CAST"kysec_node");
	xmlNewNsProp(newNode, NULL, BAD_CAST"path", BAD_CAST(path));
	xmlAddChild(curUserNode, newNode);
	hashNode = xmlNewNode(NULL, BAD_CAST"hash");
	typeNode = xmlNewNode(NULL, BAD_CAST"type");
	nameNode = xmlNewNode(NULL, BAD_CAST"name");
	icoNode = xmlNewNode(NULL, BAD_CAST"ico");
	serverTypeNode = xmlNewNode(NULL, BAD_CAST"server_type");

	/* hash */
	xmlAddChild(newNode, hashNode);
	hashVal = get_path_hash(path, uid);
	xmlAddChild(hashNode, xmlNewText(BAD_CAST(hashVal)));

	/* type */
	xmlAddChild(newNode, typeNode);
	memset(type_tmp, '\0', sizeof(type_tmp));
	snprintf(type_tmp, sizeof(type_tmp), "%d", type);
	xmlAddChild(typeNode, xmlNewText(BAD_CAST(type_tmp)));

	/* name */
	xmlAddChild(newNode, nameNode);
	xmlAddChild(nameNode, xmlNewText(BAD_CAST(name)));

	/* ico */
	xmlAddChild(newNode, icoNode);
	xmlAddChild(icoNode, xmlNewText(BAD_CAST(ico)));

	/* server type */
	xmlAddChild(newNode, serverTypeNode);
	memset(server_type_tmp, '\0', sizeof(server_type_tmp));
	snprintf(server_type_tmp, sizeof(server_type_tmp), "%d", server_type);
	xmlAddChild(serverTypeNode, xmlNewText(BAD_CAST(server_type_tmp)));

	ret = xmlSaveFile(KYSEC_XML_PATH, doc);
	if (ret >= 0)
		ret = 0;

	if (hashVal)
		free(hashVal);
out:
	xmlFreeDoc(doc);
	return ret;
}

/* return: 0 is success, -1 is error */
int kysec_xml_remove_node(uid_t uid, char *path)
{
	xmlDocPtr doc;
	xmlNodePtr rootNode = NULL;
	xmlNodePtr curNode = NULL;
	xmlNodePtr userNode = NULL;
	xmlNodePtr curUserNode = NULL;
	struct kysec_xml *xmlData = NULL;
	int ret = 0;

	if (!path)
		return -1;

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return -1;
	}

	/* get root */
	rootNode = xmlDocGetRootElement(doc);
	if (!rootNode) {
		xmlFreeDoc(doc);
		return -1;
	}

	if (xmlStrcmp(rootNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		xmlFreeDoc(doc);
		return -1;
	}

	/* kysec_root children */
	userNode = rootNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		if (!userID)
			goto user_next;
		if (userID && uid != atoi(userID)) {
			xmlFree(userID);
			goto user_next;
		}

		xmlFree(userID);
		/* find user node */
		curUserNode = userNode;
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlAttrPtr attr;
				xmlChar *path_tmp = NULL;

				if (!xmlHasProp(curNode,BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path_tmp = xmlGetProp(curNode, BAD_CAST"path");
				/* exist */
				if (path_tmp && !xmlStrcmp(path_tmp, BAD_CAST(path))) {
					xmlNodePtr tempNode;

					/* remove 'kysec_node' */
					tempNode = curNode->next;
					xmlUnlinkNode(curNode);
					xmlFreeNode(curNode);
					curNode = tempNode;

					xmlFree(path_tmp);
					ret = 0;
					goto out;
				}
				if (path_tmp)
					xmlFree(path_tmp);
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		/* have found user node, not need to loop */
		if (curUserNode)
			break;
		userNode = userNode->next;
	}
out:
	ret = xmlSaveFile(KYSEC_XML_PATH, doc);
	if (ret >= 0)
		ret = 0;

	xmlFreeDoc(doc);
	return ret;
}

/* return: 0 is success, -1 is error */
int kysec_xml_update_node(uid_t uid, char *path, int type,
		char *name, char *ico, int server_type)
{
	xmlDocPtr doc;
	xmlNodePtr rootNode = NULL;
	xmlNodePtr curNode = NULL;
	xmlNodePtr userNode = NULL;
	xmlNodePtr curUserNode = NULL;
	int ret = 0;
	char type_tmp[10] = {0};
	char server_type_tmp[10] = {0};

	if (!path)
		return -1;

	if (type < OP_ALLOW || type > OP_AUTH) {
		printf("The type (%d) is a wrong operation type\n", type);
		return -1;
	}

	xmlResetLastError();
	doc = xmlReadFile(KYSEC_XML_PATH, "UTF-8", XML_PARSE_RECOVER);
	if (!doc) {
		kysec_xml_error_handle();
		return -1;
	}

	/* get root */
	rootNode = xmlDocGetRootElement(doc);
	if (!rootNode) {
		xmlFreeDoc(doc);
		return -1;
	}

	if (xmlStrcmp(rootNode->name, BAD_CAST"kysec_root")) {
		printf("document of the wrong type, root node != kysec_root\n");
		xmlFreeDoc(doc);
		return -1;
	}

	/* kysec_root children */
	userNode = rootNode->xmlChildrenNode;
	while (userNode != NULL) {
		if (xmlStrcmp(userNode->name, (const xmlChar *)"kysec_user"))
			goto user_next;

		if (!xmlHasProp(userNode, BAD_CAST"userID"))
			goto user_next;

		xmlAttrPtr userAttr;
		xmlChar *userID = NULL;

		userAttr = userNode->properties;
		if (xmlStrcmp(userAttr->name, BAD_CAST"userID"))
			goto user_next;
		userID = xmlGetProp(userNode, BAD_CAST"userID");
		if (!userID)
			goto user_next;
		if (userID && uid != atoi(userID)) {
			xmlFree(userID);
			goto user_next;
		}

		xmlFree(userID);
		/* find user node */
		curUserNode = userNode;
		curNode = userNode->xmlChildrenNode;
		while (curNode != NULL) {
			if (!xmlStrcmp(curNode->name, (const xmlChar *)"kysec_node")) {
				xmlNodePtr childNode;
				xmlAttrPtr attr;
				xmlChar *path_tmp = NULL;

				if (!xmlHasProp(curNode,BAD_CAST"path"))
					goto cur_next;

				attr = curNode->properties;
				if (xmlStrcmp(attr->name, BAD_CAST"path"))
					goto cur_next;
				path_tmp = xmlGetProp(curNode, BAD_CAST"path");
				if (path_tmp && xmlStrcmp(path_tmp, BAD_CAST(path))) {
					xmlFree(path_tmp);
					goto cur_next;
				}

				/* found, kysec_node children */
				childNode = curNode->xmlChildrenNode;
				while (childNode != NULL) {
					if (!xmlStrcmp(childNode->name, (const xmlChar *)"type")) {
						if (type == OP_UNKNOWN)
							goto child_next;
						/* update 'type' */
						memset(type_tmp, '\0', sizeof(type_tmp));
						snprintf(type_tmp, sizeof(type_tmp), "%d", type);
						xmlNodeSetContent(childNode, BAD_CAST(type_tmp));
					} else if (!xmlStrcmp(childNode->name, (const xmlChar *)"name")) {
						if (!name)
							goto child_next;
						/* update 'name' */
						xmlNodeSetContent(childNode, BAD_CAST(name));
					} else if (!xmlStrcmp(childNode->name, (const xmlChar *)"ico")) {
						if (!ico)
							goto child_next;
						/* update 'ico' */
						xmlNodeSetContent(childNode, BAD_CAST(ico));
					} else if (!xmlStrcmp(childNode->name, (const xmlChar *)"server_type")) {
						if (server_type == SERV_UNKNOWN)
							goto child_next;
						/* update 'server_type' */
						memset(server_type_tmp, '\0', sizeof(server_type_tmp));
						snprintf(server_type_tmp, sizeof(server_type_tmp), "%d", server_type);
						xmlNodeSetContent(childNode, BAD_CAST(server_type_tmp));
					}
child_next:
					childNode = childNode->next;
				}
				if (path_tmp)
					xmlFree(path_tmp);
			}
cur_next:
			curNode = curNode->next;
		}
user_next:
		/* have found user node, not need to loop */
		if (curUserNode)
			break;
		userNode = userNode->next;
	}
out:
	ret = xmlSaveFile(KYSEC_XML_PATH, doc);
	if (ret >= 0)
		ret = 0;

	xmlFreeDoc(doc);
	return ret;
}

#if 0
int main(int argc, char *argv[])
{
	int ret = 0, i = 0;
	int count = 0;
	struct kysec_xml *data = NULL;
	char *cmd = NULL;
	char *filepath = NULL;
	int op_type = OP_ALLOW;
	uid_t uid;

	if (argc < 2) {
		printf("Usage: %s [ add | del | update | lookup ] [filepath] [op_type] uid\n\n", argv[0]);
		printf("%s add     filepath op_type uid\n", argv[0]);
		printf("%s update  filepath op_type uid\n", argv[0]);
		printf("%s del     filepath uid\n", argv[0]);
		printf("%s lookup [uid]\n", argv[0]);
		return -1;
	}

	cmd = argv[1];

	if (strcmp(cmd, "add") == 0 && argc == 5) {
		filepath = argv[2];
		op_type = atoi(argv[3]);
		uid = atoi(argv[4]);
		ret = kysec_xml_add_node(uid, filepath, op_type);
		printf("add ret = %d\n", ret);
	} else if (strcmp(cmd, "lookup") == 0) {
		if (argc > 2) {
			uid = atoi(argv[2]);
			data = kysec_xml_read_all_data_with_uid(uid, &count);
		} else
			data = kysec_xml_read_all_data(&count);
		printf("count=%d\n", count);
		for (i = 0; i < count; i++)
			printf("%s : %s : %d\n", data[i].path, data[i].hash, data[i].type);
		free(data);
	} else if (strcmp(cmd, "update") == 0 && argc == 5) {
		filepath = argv[2];
		op_type = atoi(argv[3]);
		uid = atoi(argv[4]);
		ret = kysec_xml_update_node(uid, filepath, op_type);
		printf("update ret = %d\n", ret);
	} else if (strcmp(cmd, "del") == 0 && argc == 4) {
		filepath = argv[2];
		uid = atoi(argv[3]);
		ret = kysec_xml_remove_node(uid, filepath);
		printf("remove ret = %d\n", ret);
	} else {
		printf("The arguments wrong\n");
		return -1;
	}

	return 0;
}
#endif
