/*
 *  Kylin security execute control libraries
 *
 *  This file contains the Kylin security exectl label manage
 *  implementations.
 *
 *  Authors:  songchuanfei <songchuanfei@kylinos.cn>
 *  Authors:  huanfuxi <huanfuxi@kylinos.cn>
 *  Authors:  mengyuan <mengyuan@kylinos.cn>
 *
 *  Copyright (C) 2017,Tianjin KYLIN Information Technology Co., Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */
#include <kysec/libkysec.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <limits.h>
#include <libsecurity.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#define KYSEC_LABEL_MAX 255

enum {
	AUTH_ERROR	= -1,		//认证出错
	AUTH_FAILED	=  0,		//认证失败
	AUTH_SUCCESS	=  1,		//认证成功
	AUTH_SUCCESS_ONCE = 2,
	AUTH_SUCCESS_DENY = 3
};

enum {
	AUTH_KERNEL = 0,
	AUTH_USER
};

enum authorized_type {
	AUTH_TYPE_EXECTL = 1,
	AUTH_TYPE_NETCTL
};

struct auth_data {
	int auth_type;
	uid_t uid;
	pid_t pid;			//待认证的进程PID
	uint64_t token;			//认证令牌
	char filepath[PATH_MAX + 1];	//待认证的操作对象(文件)
};

#define PID_DEEP 20
struct forefather_pid {
	pid_t pids[PID_DEEP];           /* forefather pid */
	int auth;                       /* forefather auth result */
	int count;                      /* forefather pid counter */
};

struct auth_result {
	struct forefather_pid fpid;	//祖先进程PID信息
	pid_t pid;			//待认证的进程PID
	int authorized;			//认证结果
	uint64_t token;			// 认证令牌
};

#define KYSEC_AUTH_RESULT_TMP	"/tmp/kysec-auth.pid"
struct forefather_pid fpid;
uint64_t curr_token;

static void do_syslog(const char *format, ...)
{
	va_list vlist;

	va_start(vlist, format);

	openlog("KYSEC(LIB)", 0, LOG_LOCAL0);
	vsyslog(LOG_WARNING, format, vlist);
	closelog();

	va_end(vlist);
}

static int write_auth_data(struct auth_data *data)
{
	int fd = -1;
	ssize_t size = 0;
	mode_t um = 0;
	int rc = 0;

	um = umask(0);
	fd = open(KYSEC_AUTH_RESULT_TMP, O_CREAT | O_RDWR | O_TRUNC, 0666);
	if (fd < 0) {
		do_syslog("%s: failed to open file '%s': %s\n",
				__func__, KYSEC_AUTH_RESULT_TMP, strerror(errno));
		rc = -1;
		goto out;
	}

	size = write(fd, data, sizeof(struct auth_data));
	if (size <= 0) {
		do_syslog("failed to write auth data: %s\n",
				strerror(errno));
		rc = -1;
		goto out;
	}

out:
	close(fd);
	umask(um);

	return rc;
}

static int read_auth_result(struct auth_result *result)
{
	int fd = -1;
	ssize_t size = 0;

	if (NULL == result) {
		errno = EINVAL;
		do_syslog("%s: %s\n", __func__, strerror(errno));
		return -1;
	}

	fd = open(KYSEC_AUTH_RESULT_TMP, O_RDWR);
	if (fd < 0) {
		do_syslog("%s: failed to open file '%s': %s\n",
			  __func__, KYSEC_AUTH_RESULT_TMP, strerror(errno));
		return -1;
	}

	size = read(fd, result, sizeof(struct auth_result));
	if (size <= 0 || (size > 0 && size < (int)sizeof(*result))) {
		do_syslog("failed to read file '%s'\n", KYSEC_AUTH_RESULT_TMP);
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

static int write_auth_result(struct auth_result *result)
{
	int fd = -1;
	ssize_t size = 0;
	mode_t um = 0;
	int rc = 0;

	um = umask(0);
	fd = open(KYSEC_AUTH_RESULT_TMP, O_CREAT | O_RDWR | O_TRUNC, 0666);
	if (fd < 0) {
		do_syslog("%s: failed to open file '%s': %s\n",
				__func__, KYSEC_AUTH_RESULT_TMP, strerror(errno));
		rc = -1;
		goto out;
	}

	size = write(fd, result, sizeof(struct auth_result));
	if (size <= 0) {
		do_syslog("failed to write auth result: %s\n",
				strerror(errno));
		rc = -1;
		goto out;
	}

out:
	close(fd);
	umask(um);

	return rc;
}

int kysec_label_get(char *path, char **label)
{
	if (kysec_is_disabled()) {
		errno = ENOTSUP;
		return -1;
	}

	*label = calloc(sizeof(char), KYSEC_LABEL_MAX + 1);
	int rc = lgetxattr(path, XATTR_NAME_KYSEC, *label,
			   KYSEC_LABEL_MAX);

	if (rc < 0)
		free(*label);

	return rc;
}

static int set_argv_and_envp(char *argv[], char *envp[])
{
	char *home = NULL;
	char *display = NULL;
	char *xauth = NULL;
	char *tty_name = NULL;
	char xfile[512] = {0};

	home = getenv("HOME");
	if (!home)
		return -1;

	display = getenv("DISPLAY");
	if (!display) {
		tty_name = ttyname(0);
		if (!tty_name) {
			do_syslog("failed to get ttyname: %s\n", strerror(errno));
			return -1;
		}
	} else {
		xauth = getenv("XAUTHORITY");
		if (xauth)
			snprintf(xfile, sizeof(xfile), "%s", xauth);
		else
			snprintf(xfile, sizeof(xfile), "%s/.Xauthority", home);
	}

	envp[0] = home;
	envp[1] = "PATH=/sbin:/bin:/usr/sbin:/usr/bin";
	snprintf(envp[2], 1024, "DISPLAY=%s", display);
	snprintf(envp[3], 1024, "XAUTHORITY=%s", xfile);

	argv[0] = "/sbin/kysec_auth";
	snprintf(argv[1], 11, "%s", "--caller");
	snprintf(argv[2], 5, "%s", "user");
	if (!tty_name) {
		argv[3] = NULL;
		argv[4] = NULL;
	} else {
		argv[3] = "--tty";
		snprintf(argv[4], 20, "%s", tty_name);
	}
	argv[5] = NULL;

	return 0;
}

static int kysec_get_forefather_auth(pid_t ppid, pid_t pid)
{
	struct auth_result result;
	int ret, i = 0;
	int count = 0;

	memset(&result, 0, sizeof(struct auth_result));
	ret = read_auth_result(&result);
	if (ret != 0) {
		ret = AUTH_ERROR;
		goto error;
	}

#if 0
	do_syslog("KYSEC: Auth token 0x%llx, 0x%llx\n", result.token, curr_token);
	if (result.token != curr_token) {
		do_syslog("KYSEC: Auth token 0x%llx not match\n", curr_token);
		ret = AUTH_ERROR;
		goto error;
	}
#endif

	count = result.fpid.count;
	/* forefather pids over limit */
	if (count >= PID_DEEP) {
		ret = AUTH_ERROR;
		goto error;
	}
	do_syslog("KYSEC(read): forefather info: count=%d\n", count);
	for (i = 0; i < count; i++)
		do_syslog("%d ", result.fpid.pids[i]);
	do_syslog("KYSEC(read): forefather pid:(%d), current pid: %d(%d)\n",
			result.fpid.auth, result.pid, result.authorized);

	for (i = 0; i < count; i++) {
		if (ppid == result.fpid.pids[i] || pid == result.fpid.pids[i]) {
			ret = result.fpid.auth;
			goto error;
		}
	}

	if (ppid == result.pid || pid == result.pid) {
		ret = result.authorized;
		result.fpid.pids[count] = result.pid;
		result.fpid.auth = ret;
		result.fpid.count++;
	} else
		ret = AUTH_ERROR;

error:
	/* keep forefather pid and auth result unchanged */
	memset(&fpid, 0, sizeof(struct forefather_pid));
	if (ret == AUTH_ERROR) {
		fpid.pids[0] = pid;
		fpid.auth = ret;
		fpid.count = 1;
	} else
		memcpy(&fpid, &result.fpid, sizeof(struct forefather_pid));

	return ret;
}

static void kysec_record_auth_result(int auth, pid_t pid)
{
	int ret = 0, i, count;
	struct auth_result result;

	memset(&result, 0, sizeof(struct auth_result));
	memcpy(&result.fpid, &fpid, sizeof(struct forefather_pid));
	result.pid = pid;
	result.authorized = auth;
	result.token = curr_token;

	count = fpid.count;
	do_syslog("KYSEC(write): forefather info: count=%d\n", count);
	for (i = 0; i < count; i++)
		do_syslog("%d ", fpid.pids[i]);
	do_syslog("KYSEC(write): forefather pid:(%d), current pid: %d(%d)\n",
			fpid.auth, result.pid, result.authorized);

	ret = write_auth_result(&result);
	if (ret < 0)
		do_syslog("KYSEC: Failed to record auth result\n");
}

static int kysec_do_auth(char *path)
{
	int rc = 0, auth;
	char *argv[6] = {NULL};
	static char *envp[5] = {NULL};
	char absolute_path[PATH_MAX + 1] = {0};
	pid_t pid;
	pid_t ppid;
	pid_t current_pid;
	struct auth_data data;
	struct auth_result result;
	int path_len = 0;

	if (!realpath(path, absolute_path)) {
		do_syslog("failed to get '%s' realpath: %s\n", path, strerror(errno));
		return 0;
	}
	path_len = strlen(absolute_path);

	/* need authorized by manual operation */

	argv[1] = (char *)malloc(11);
	argv[2] = (char *)malloc(path_len + 1);
	argv[4] = (char *)malloc(20);
	envp[2] = (char *)malloc(1024);
	envp[3] = (char *)malloc(1024);
	if (!argv[1] || !argv[2] || !argv[4] || !envp[2] || !envp[3]) {
		rc = 0;
		goto out;
	}

	if (set_argv_and_envp(argv, envp) < 0) {
		rc = 0;
		goto out;
	}

	ppid = getppid();
	current_pid = getpid();
	do_syslog("current_pid=%d ppid=%d\n", current_pid, ppid);
	auth = kysec_get_forefather_auth(ppid, current_pid);
	if (auth == AUTH_SUCCESS) {
		rc = 1;
		snprintf(argv[1], 11, "%s", "--setlabel");
		snprintf(argv[2], path_len + 1, "%s", absolute_path);
		goto popup;
	} else if (auth == AUTH_SUCCESS_ONCE) {
		rc = 1;
		goto out;
	} else if (auth == AUTH_FAILED || auth == AUTH_SUCCESS_DENY) {
		rc = 0;
		goto out;
	}

	memset(&data, 0, sizeof(struct auth_data));
	memset(&result, 0, sizeof(struct auth_result));
	srandom(time(NULL));
	curr_token = random();
	data.auth_type = AUTH_TYPE_EXECTL;
	data.uid = geteuid();
	data.pid = current_pid;
	data.token = curr_token;
	snprintf(data.filepath, sizeof(data.filepath), "%s", absolute_path);

	if (write_auth_data(&data) < 0) {
		rc = 0;
		goto out;
	}

popup:
	pid = fork();
	if (0 == pid) {
		if (execvpe(argv[0], argv, envp) < 0)
			rc = 0;
	} else if (pid > 0) {
		waitpid(pid, NULL, 0);

		if (auth == AUTH_SUCCESS)
			goto out;

		if (read_auth_result(&result) == 0) {
			if (current_pid == result.pid &&
			    curr_token == result.token &&
			    result.authorized >= AUTH_SUCCESS) {
				auth = result.authorized;
				if (auth == AUTH_SUCCESS_DENY)
					rc = 0;
				else
					rc = 1;
			}
		}
	}

out:
	free(argv[1]);
	free(argv[2]);
	free(argv[4]);
	free(envp[2]);
	free(envp[3]);

	if (fpid.auth == AUTH_ERROR)
		fpid.auth = auth;
	kysec_record_auth_result(auth, current_pid);

	return rc;
}

int kysec_check_exec(char *path)
{
	char *exectl_sid = NULL;
	int status, exectl_status, retval = 1;
	int len = 0;

	if (kysec_is_disabled())
		return 1;

	exectl_status = kysec_get_func_status(FUNC_EXECTL);
	if (exectl_status == FUNC_DISABLED)
		return 1;

	status = kysec_getstatus();
	if (status == KYSEC_STATUS_SOFTMODE)
		return 1;

	if (kysec_exectl_get(path, &exectl_sid) < 0) {
		if (errno == ENOTSUP ||
		    errno == ENODATA ||
		    errno == ENOENT)
			retval = 1;
		else
			retval = 0;
		goto out;
	}

	if (!strcmp(exectl_sid, "unknown")) {
		if (status == KYSEC_STATUS_NORMAL)
			retval = 0;
		else if (status == KYSEC_STATUS_WARNING)
			retval = kysec_do_auth(path);
	}

out:
	free(exectl_sid);
	return retval;
}

int kysec_check_kmod(char *path)
{
	char *exectl_sid = NULL;
	int status, exectl_status, retval = 1;
	int len = 0;

	if (kysec_is_disabled())
		return 1;

	exectl_status = kysec_get_func_status(FUNC_KMOD);
	if (exectl_status == FUNC_DISABLED)
		return 1;

	status = kysec_getstatus();
	if (status == KYSEC_STATUS_SOFTMODE)
		return 1;

	if (kysec_exectl_get(path, &exectl_sid) < 0) {
		if (errno == ENOTSUP ||
		    errno == ENODATA ||
		    errno == ENOENT)
			retval = 1;
		else
			retval = 0;
		goto out;
	}

	if (!strcmp(exectl_sid, "unknown")) {
		if (status == KYSEC_STATUS_NORMAL)
			retval = 0;
		else if (status == KYSEC_STATUS_WARNING)
			retval = kysec_do_auth(path);
	}

out:
	free(exectl_sid);
	return retval;
}

int kysec_is_disabled(void)
{
	return (kysec_getstatus() == KYSEC_STATUS_DISABLED);
}

int kysec_label_get_by_pid(pid_t pid, char **label)
{
	char buffer[KYSEC_LABEL_MAX+1] = {0};
	char *sid_path = NULL;
	int fd = -1, retval = 0;

	if (kysec_is_disabled()) {
		errno = ENOTSUP;
		return -1;
	}

	sid_path = calloc(sizeof(char), PATH_MAX);
	if (!sid_path)
		return -1;

	if (security_lsm_is_enabled())
		sprintf(sid_path, "/proc/%d/attr/kysec/current", pid);
	else
		sprintf(sid_path, "/proc/%d/attr/current", pid);

	fd = open(sid_path, O_RDONLY);
	if (fd < 0) {
		free(sid_path);
		return -1;
	}

	retval = read(fd, buffer, sizeof(buffer) - 1);

	if (retval > 0) {
		buffer[retval] = 0;
		*label = strdup(buffer);
		if (!label)
			retval = -1;
		else
			retval = 0;
	} else
		retval = -1;

out:
	if (sid_path)
		free(sid_path);
	close(fd);
	return retval;
}

int kysec_exectl_get(char *path, char **label)
{
	char *kysec_sid, *pos;
	int retval = 0;

	*label = NULL;
	if (kysec_label_get(path, &kysec_sid) < 0)
		return -1;

	pos = strrchr(kysec_sid, ':');
	if (!pos) {
		errno = EINVAL;
		retval = -1;
	} else {
		*label = strdup(pos+1);
		if (!*label)
			retval = -1;
	}

	free(kysec_sid);
	return retval;
}

int kysec_protect_get(char *path, char **label)
{
	char *kysec_sid, *pos;
	int retval = 0;

	*label = NULL;
	if (kysec_label_get(path, &kysec_sid) < 0)
		return -1;

	pos = strchr(kysec_sid, ':');
	if (!pos) {
		errno = EINVAL;
		retval = -1;
	} else {
		char *end = strrchr(kysec_sid, ':');

		if (!end) {
			errno = EINVAL;
			retval = -1;
		} else {
			*end = '\0';
			*label = strdup(pos+1);
			if (!*label)
				retval = -1;
		}
	}

	free(kysec_sid);
	return retval;
}
int kysec_identify_get(char *path, char **label)
{
	char *kysec_sid, *pos;
	int retval = 0;

	*label = NULL;
	if (kysec_label_get(path, &kysec_sid) < 0)
		return -1;

	pos = strchr(kysec_sid, ':');
	if (!pos) {
		errno = EINVAL;
		retval = -1;
	} else {
		*pos = '\0';
		*label = strdup(kysec_sid);
		if (!*label)
			retval = -1;
	}

	free(kysec_sid);
	return retval;
}
