/*
 *  Kylin security execute control libraries
 *
 *  This file contains the Kylin security exectl label manage
 *  implementations.
 *
 *  Authors:  songchuanfei <songchuanfei@kylinos.cn>
 *  Authors:  huanfuxi <huanfuxi@kylinos.cn>
 *  Authors:  mengyuan <mengyuan@kylinos.cn>
 *
 *  Copyright (C) 2017,Tianjin KYLIN Information Technology Co., Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */
#include <kysec/libkysec_dbus.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <limits.h>
#include <libsecurity.h>
#include <glib.h>
#include <gio/gio.h>
#include <locale.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <stdint.h>
#include <kysec/kysec_communicate.h>
#include <syslog.h>

#define SYSTEM_DBUS_SERVICE		"com.kylin.kysec"
#define SYSTEM_OBJECT_PATH		"/xattr"
#define SYSTEM_DBUS_INTERFACE	"com.kylin.kysec.xattr"
#define SYSTEM_DBUS_TIMEOUT		5000

static char dbus_err_msg[2048] = {0};

static void do_syslog(const char *format, ...)
{
	va_list vlist;

	va_start(vlist, format);

	openlog("KYSEC_DBUS(LIB)", 0, LOG_LOCAL0);
	vsyslog(LOG_WARNING, format, vlist);
	closelog();

	va_end(vlist);
}

int handling_dbus_error(GError *error)
{
	int retval = 0;
	gchar *mesg = NULL;
	gchar *errName = NULL;

	if (g_dbus_error_is_remote_error(error)) {
		errName = g_dbus_error_get_remote_error(error);
		g_dbus_error_strip_remote_error(error);

		mesg = error->message;
		if (mesg) {
			do_syslog("%s: %s\n", errName, mesg);
			memset(dbus_err_msg, 0x00, sizeof(dbus_err_msg));
			snprintf(dbus_err_msg, sizeof(dbus_err_msg), "%s", mesg);
		}
		
		if (strcmp(errName, "KYLIN.KYSEC.OK") == 0)
			retval = 0;
		else if (strcmp(errName, "KYLIN.KYSEC.AUTH.FAILED") == 0)
			retval = 1;
		else if (strcmp(errName, "KYLIN.KYSEC.CANCEL.AUTH") == 0)
			retval = 2;
		else if (strcmp(errName, "KYLIN.KYSEC.DISABLED") == 0)
			retval = 3;
		else if (strcmp(errName, "KYLIN.KYSEC.DATABASE.ERR") == 0)
			retval = 4;
		else if (strcmp(errName, "KYLIN.KYSEC.LABEL.GET.ERR") == 0)
			retval = 5;
		else if (strcmp(errName, "KYLIN.KYSEC.LABEL.SET.ERR") == 0)
			retval = 6;
		else if (strcmp(errName, "KYLIN.KYSEC.INVALID.ARG") == 0)
			retval = 7;
		else if (strcmp(errName, "KYLIN.KYSEC.EXTRACT.LIST.ERR") == 0)
			retval = 8;
		else if (strcmp(errName, "KYLIN.KYSEC.FILE.ATTR.ERR") == 0)
			retval = 9;
		else 
			retval = 10;		
	}

	return retval;
}

GVariant *process_dbus_method(gchar *method, GVariant *parameters, int *err_code)
{
	if (NULL == method) {
		do_syslog("%s: the method is NULL\n", __func__);
		*err_code = 8;
		return NULL;
	}

	if (NULL == parameters) {
		do_syslog("%s: the parameter is NULL\n", __func__);
		*err_code = 8;
		return NULL;
	}

	GError *error = NULL;
	GDBusProxy *proxy = NULL;
	GVariant *result = NULL;

	proxy = g_dbus_proxy_new_for_bus_sync(
				G_BUS_TYPE_SYSTEM,
				G_DBUS_PROXY_FLAGS_NONE,
				NULL,
				SYSTEM_DBUS_SERVICE,
				SYSTEM_OBJECT_PATH,
				SYSTEM_DBUS_INTERFACE,
				NULL,
				&error
				);

	if (error) {
		*err_code = handling_dbus_error(error);
		g_error_free(error);
		error = NULL;
		return NULL;
	}

	result = g_dbus_proxy_call_sync(
				proxy,
				method,
				parameters,
				G_DBUS_CALL_FLAGS_NONE,
				G_MAXINT,
				NULL,
				&error);

	if (error) {
		*err_code = handling_dbus_error(error);
		g_error_free(error);
		error = NULL;
		g_object_unref(proxy);
		return NULL;
	}

	g_object_unref(proxy);
	return result;
}

static int path_is_dir(const char *path)
{
	struct stat sb;

	if (!lstat(path, &sb) && S_ISDIR(sb.st_mode))
		return 1;

	return 0;
}

char* get_full_path(char* fpath)
{
	char* full_path = NULL;
	char real_path[PATH_MAX] = {0};	

	if (access(fpath, F_OK) != 0)	{
		full_path = strdup(fpath);
		return full_path;
	}

	if (path_is_dir(fpath))	{
		realpath(fpath, real_path);
		full_path = strdup(real_path);
		return full_path;
	}

	char path[PATH_MAX] = {0};
	strncpy(path, fpath, sizeof(path));
	char* p = strrchr(path, '/');
	if (p != NULL)	{
		char* file_name = p + 1;
		*p = '\0';

		realpath(path, real_path);

		char apath[PATH_MAX] = {0};
		snprintf(apath, sizeof(apath), "%s/%s", real_path, file_name);
		full_path = strdup(apath);
	}
	else	{
		char pwd[PATH_MAX] = {0};
		getcwd(pwd, sizeof(pwd));
		char apath[PATH_MAX] = {0};
		snprintf(apath, sizeof(apath), "%s/%s", pwd, fpath);
		full_path = strdup(apath);
	}
	return full_path;
}

static char *get_file_hash_value(char *path)
{
	char *hash = NULL;
	GVariant *parameters = g_variant_new("(s)", path);
	gchar *ghash = NULL;

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_get_hash_by_path", parameters, &err_code);
	if (NULL == result)
		return NULL;

	ghash = g_variant_get_string(g_variant_get_child_value(result, 0), NULL);

	if (strlen(ghash) != 0)
		hash = strdup(ghash);

	g_variant_unref(result);
	return hash;
}

static char *get_exectl_label_from_kysec_label(char *kysec_label)
{
	if (NULL == kysec_label)
		return NULL;

	char *p;
	/*strtok places a NULL terminator infront of the token, if found */
	p = strtok(kysec_label, ":");
	p = strtok(NULL, ",");
	p = strtok(NULL, ",");

	return p;
}

static int process_kernel_hash_value(char *exectl, char *hash)
{
	int operate = 1;
	int mesg_type;

	if (strcmp(exectl, "unknown") != 0) {
		mesg_type = LABEL_ADD;
		operate = kysec_convert_exectl_label_to_label_type(exectl);
	} else
		mesg_type = LABEL_DEL;

	int retval = kysec_hash_communicate_with_kernel(mesg_type, hash, operate);
	if (retval != 0) {
		do_syslog("kysec hash communicate with kernel failed, retval = %d\n", retval);
		return -1;
	}

	return 0;
}

char *get_dbus_err_msg()
{
	return dbus_err_msg;
}

int kysec_label_set(char *path, char *label)
{
	char *exectl_label = NULL;
	char *hash = NULL;
	int retval = 0;
	int ret = 0;

	if (path == NULL || label == NULL) {
		errno = EINVAL;
		return -2;
	}

	if (kysec_is_disabled()) {
		errno = ENOTSUP;
		return -1;
	}

	char* absolute_path = get_full_path(path);
	char *old_exectl;
	if (0 != kysec_exectl_get(absolute_path, &old_exectl))	{
		errno = EINVAL;
		free(absolute_path);
		return -1;
	}

	exectl_label = get_exectl_label_from_kysec_label(label);
	if (NULL == exectl_label) {
		errno = EINVAL;
		free(absolute_path);
		free(old_exectl);
		do_syslog("get exectl label from kysec label failed, kysec label = %s\n", label);
		return -3;
	}

	hash = get_file_hash_value(absolute_path);
	if (NULL != hash) {		
		ret = process_kernel_hash_value("unknown", hash);
		if (ret != 0) {
			memset(dbus_err_msg, 0x00, sizeof(dbus_err_msg));
			snprintf(dbus_err_msg, sizeof(dbus_err_msg), "set hash to kernel failed");
			do_syslog("process_kernel_hash_value failed, file = %s, label = %s， hash = %s\n",
					absolute_path, exectl_label, hash);
			free(hash);
			free(old_exectl);
			free(absolute_path);
			return -4;
		}
		free(hash);
	}

	GVariant *g_label = g_variant_new("s", label);
	GVariantBuilder *g_builder_parameter = g_variant_builder_new(G_VARIANT_TYPE_TUPLE);
	GVariantBuilder *g_builder_file_list = g_variant_builder_new(G_VARIANT_TYPE("as"));
	g_variant_builder_add(g_builder_file_list, "s", absolute_path);
	GVariant *g_file_list = g_variant_builder_end(g_builder_file_list);
	g_variant_builder_add_value(g_builder_parameter, g_file_list);
	g_variant_builder_add_value(g_builder_parameter, g_label);
	GVariant *parameters = g_variant_builder_end(g_builder_parameter);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_label_set", parameters, &err_code);
	if (NULL == result) {
		ret = err_code;
		goto out;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));
	g_variant_unref(result);	

out:
	g_variant_builder_unref(g_builder_file_list);
	g_variant_builder_unref(g_builder_parameter);

	hash = get_file_hash_value(absolute_path);
	if (hash != NULL) {
		if (ret == 0)	{
			if (strcmp(exectl_label, "unknown") != 0)
				ret = process_kernel_hash_value(exectl_label, hash);
		}
		else
			ret = process_kernel_hash_value(old_exectl, hash);
		if (ret != 0) {
			memset(dbus_err_msg, 0x00, sizeof(dbus_err_msg));
			snprintf(dbus_err_msg, sizeof(dbus_err_msg), "set hash to kernel failed");
			do_syslog("process_kernel_hash_value failed, file = %s, label = %s， hash = %s\n",
						absolute_path, exectl_label, hash);
		}
		free(hash);			
	}	

	free(absolute_path);
	free(old_exectl);

	return ret;
}

int kysec_protect_set(char *path, char *protect)
{
	int retval = -1;
	int ret = 0;

	if (path == NULL || protect == NULL) {
		errno = EINVAL;
		return -2;
	}

	if (!strcmp(protect, "readonly") ||
	    !strcmp(protect, "verify") ||
	    !strcmp(protect, "none"))
		retval = 0;

	if (retval < 0) {
		errno = EINVAL;
		return -1;
	}

	char *absolute_path = get_full_path(path);
	GVariant *g_label = g_variant_new("s", protect);

	GVariantBuilder *g_builder_parameter = g_variant_builder_new(G_VARIANT_TYPE_TUPLE);

	GVariantBuilder *g_builder_file_list = g_variant_builder_new(G_VARIANT_TYPE("as"));
	g_variant_builder_add(g_builder_file_list, "s", absolute_path);

	GVariant *g_file_list = g_variant_builder_end(g_builder_file_list);

	g_variant_builder_add_value(g_builder_parameter, g_file_list);
	g_variant_builder_add_value(g_builder_parameter, g_label);

	GVariant *parameters = g_variant_builder_end(g_builder_parameter);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_protect_set", parameters, &err_code);
	if (NULL == result) {
		ret = err_code;
		goto out;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));

	g_variant_unref(result);
out:
	g_variant_builder_unref(g_builder_file_list);
	g_variant_builder_unref(g_builder_parameter);
	free(absolute_path);

	return ret;
}

int kysec_identify_set(char *path, char *identify)
{
	int retval = -1;
	int ret = 0;

	if (path == NULL || identify == NULL) {
		errno = EINVAL;
		return -2;
	}

	if (!strcmp(identify, "secadm") ||
	    !strcmp(identify, "audadm") ||
	    !strcmp(identify, "none"))
		retval = 0;

	if (retval < 0) {
		errno = EINVAL;
		return retval;
	}

	char *absolute_path = get_full_path(path);

	GVariant *g_label = g_variant_new("s", identify);

	GVariantBuilder *g_builder_parameter = g_variant_builder_new(G_VARIANT_TYPE_TUPLE);

	GVariantBuilder *g_builder_file_list = g_variant_builder_new(G_VARIANT_TYPE("as"));
	g_variant_builder_add(g_builder_file_list, "s", absolute_path);

	GVariant *g_file_list = g_variant_builder_end(g_builder_file_list);

	g_variant_builder_add_value(g_builder_parameter, g_file_list);
	g_variant_builder_add_value(g_builder_parameter, g_label);

	GVariant *parameters = g_variant_builder_end(g_builder_parameter);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_identify_set", parameters, &err_code);
	if (NULL == result) {
		ret = err_code;
		goto out;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));

	g_variant_unref(result);

out:
	g_variant_builder_unref(g_builder_file_list);
	g_variant_builder_unref(g_builder_parameter);
	free(absolute_path);

	return ret;
}

int kysec_exectl_set(char *path, char *exectl)
{
	char *hash = NULL;
	int retval = -1;
	int ret = 0;

	if (path == NULL || exectl == NULL) {
		errno = EINVAL;
		return -2;
	}

	if (!strcmp(exectl, "unknown") ||
	    !strcmp(exectl, "original") ||
	    !strcmp(exectl, "verified") ||
	    !strcmp(exectl, "kysoft") ||
	    !strcmp(exectl, "trusted"))
		retval = 0;

	if (retval < 0) {
		errno = EINVAL;
		return -1;
	}

	char *absolute_path = get_full_path(path);
	char *old_exectl;
	if (0 != kysec_exectl_get(absolute_path, &old_exectl))	{
		errno = EINVAL;
		free(absolute_path);
		return -1;
	}
	
	hash = get_file_hash_value(absolute_path);
	if (NULL != hash) {	//数据库中有文件hash值， 先删除内核中hash值
		ret = process_kernel_hash_value("unknown", hash);
		if (ret != 0) {
			memset(dbus_err_msg, 0x00, sizeof(dbus_err_msg));
			snprintf(dbus_err_msg, sizeof(dbus_err_msg), "set hash to kernel failed");
			do_syslog("process_kernel_hash_value failed, file = %s, label = %s， hash = %s\n",
					absolute_path, exectl, hash);
			ret = -4;
			free(hash);
			goto err_out;
		}
		free(hash);		
	}	

	GVariant *g_label = g_variant_new("s", exectl);

	GVariantBuilder *g_builder_parameter = g_variant_builder_new(G_VARIANT_TYPE_TUPLE);

	GVariantBuilder *g_builder_file_list = g_variant_builder_new(G_VARIANT_TYPE("as"));
	g_variant_builder_add(g_builder_file_list, "s", absolute_path);

	GVariant *g_file_list = g_variant_builder_end(g_builder_file_list);

	g_variant_builder_add_value(g_builder_parameter, g_file_list);
	g_variant_builder_add_value(g_builder_parameter, g_label);

	GVariant *parameters = g_variant_builder_end(g_builder_parameter);

	int err_code;
	GVariant *result = process_dbus_method("kysec_xattr_exectl_set", parameters, &err_code);
	if (NULL == result) {
		ret = err_code;
		goto out;
	}

	ret = g_variant_get_int32(g_variant_get_child_value(result, 0));
	g_variant_unref(result);
out:
	g_variant_builder_unref(g_builder_file_list);
	g_variant_builder_unref(g_builder_parameter);
	
	hash = get_file_hash_value(absolute_path);
	if (hash != NULL)	{
		if (ret == 0)	{
			if (strcmp(exectl, "unknown") != 0)
				ret = process_kernel_hash_value(exectl, hash);
		}
		else
			ret = process_kernel_hash_value(old_exectl, hash);
		if (ret != 0) {
			memset(dbus_err_msg, 0x00, sizeof(dbus_err_msg));
			snprintf(dbus_err_msg, sizeof(dbus_err_msg), "set hash to kernel failed");
			do_syslog("process_kernel_hash_value failed, file = %s, label = %s， hash = %s\n",
						absolute_path, exectl, hash);			
		}
		free(hash);		
	}

err_out:
	free(absolute_path);
	free(old_exectl);

	return ret;
}


int kysec_check_user_valid(const char *uname, const char *pwd)
{
	if (uname == NULL || pwd == NULL) {
		errno = EINVAL;
		return -2;
	}
	
	GVariant *parameters = g_variant_new("(ss)", uname, pwd);

	int err_code;
	GVariant *result = process_dbus_method("kysec_check_user_valid_with_pam", parameters, &err_code);
	if (NULL == result)
		return err_code;

	int ret = g_variant_get_int32(g_variant_get_child_value(result, 0));

	g_variant_unref(result);

	return ret;
}
