/*
 *  Kylin security execute control utils
 *
 *  Authors:  songchuanfei <songchuanfei@kylinos.cn>
 *  Authors:  huanfuxi <huanfuxi@kylinos.cn>
 *  Authors:  mengyuan <mengyuan@kylinos.cn>
 *
 *  Copyright (C) 2017,Tianjin KYLIN Information Technology Co., Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fts.h>
#include <kysec/libkysec_dbus.h>
#include <getopt.h>

#define PATH_MAX 1024

static int recursive = 0;
static int clear = 0;
static int listfile = 0;

typedef struct P_Node
{
	char *fpath;
	struct P_Node *next;
} P_Node;

//打印链表 
void print_list(P_Node *list) 
{ 
	for (P_Node *cur = list; cur != NULL; cur = cur->next) 
		printf("%s\n", cur->fpath); 
}

//申请新节点 
static P_Node* create_node(char* fpath) 
{
	if (fpath == NULL)
		return NULL;

	P_Node *node = (P_Node*)malloc(sizeof(P_Node)); 
	node->fpath = strdup(fpath); 
	node->next = NULL; 
	return node; 
} 
	
// 尾部插入 
void list_push_back(P_Node **list, char* fpath) 
{ 
	if (list == NULL)
		return;

	P_Node *node = create_node(fpath); 

	if (*list == NULL)	{
		*list = node;
		return;
	}

	//找链表中的最后一个节点 
	P_Node* cur = (*list); 
	while (cur->next != NULL) { 
		cur = cur->next; 
	} 

	cur->next = node;//插入新申请的节点 } 
	return;
}

// 销毁 ，需要销毁每一个节点 
void destory_list(P_Node **list) 
{ 
	P_Node *cur = NULL; 
	P_Node *del = NULL; 
	if (list == NULL)
		return; 

	cur = *list; 
	while (cur) { 
		del = cur; 
		cur = cur->next; 
		if (del->fpath)
			free(del->fpath);
		free(del); 
		del = NULL; 
	} 

	*list = NULL;
}

static void usage(void)
{
	printf("Usage:kysec_set [ -r ] -n LABEL_NAME -v LABEL_VALUE  PATH1 PATH2 ...\n");
	printf("      kysec_set -n LABEL_NAME -v LABEL_VALUE  -f FILE_PATH\n");
	printf("      kysec_set -n LABEL_NAME -v LABEL_VALUE  --package PACKAGE_NAME\n");
	printf("      kysec_set -n LABEL_NAME -x PATH1 PATH2 ...\n");
	printf("\t -r: set kysec label recursively, work for directories only\n");
	printf("\t -n: set the specified label type only, it can be exectl, protect or userid\n");
	printf("\t -v: new label value, label_value veris depends on label_name\n");
	printf("\t -f: set kysec label for files listed in FILE_PATH, one file path per line\n");
	printf("\t --package: set all the files' kysec labels for a installed package\n");
	printf("\t -x: clear kysec label specified by label_name\n");
	printf("\t -h: show help information.\n");
	exit(1);
}

static int kysec_set_label_for_file(char *path, int label_type, char *new_label, int set)
{
	int ret = 0;

	switch (label_type)
	{
	case 'e':
		ret = kysec_exectl_set(path, new_label);
		break;
	case 'p':
		ret = kysec_protect_set(path, new_label);
		break;
	case 'i':
		ret = kysec_identify_set(path, new_label);
		break;
	default:
		errno = EINVAL;
		ret = -1;
	}

	return ret;
}

static int check_path(const char *path)
{
	if ((access(path, F_OK)) != -1)
		return 1;
	else
		return 0;
}

static int path_is_dir(const char *path)
{
	struct stat sb;

	if (!lstat(path, &sb) && S_ISDIR(sb.st_mode))
		return 1;

	// make kernel to check whether it's a directory
	return 0;
}

static int kysec_get_file_list_from_path(P_Node **list, char* path)
{
	struct stat sb;

	if (!lstat(path, &sb) && S_ISDIR(sb.st_mode))	{
		char *pathv[2] = {0};
		if (!recursive)	{
			errno = EISDIR;
			return -1;
		}

		pathv[0] = path;
		FTS *fts = fts_open(pathv, FTS_DEFAULT, NULL);
		FTSENT *ent = NULL;

		while (1){
			ent = fts_read(fts);
			if (!ent)	{
				if (errno != 0)
					continue;
				else
					break;
			}

			if (path_is_dir(ent->fts_path))	{				
				continue;
			}
			
			list_push_back(list, ent->fts_path);
		}

		fts_close(fts);
	}
	else	{
		list_push_back(list, path);
	}

	return 0;
}

static int kysec_get_file_list_from_listfile(P_Node **list, char* path)
{
	FILE *fin;
	char *one_line;

	one_line = (char *)malloc(PATH_MAX * sizeof(char));
	if ((fin = fopen(path, "r")) == NULL) {
		free(one_line);
		return -1;
	}

	while (fgets(one_line, PATH_MAX, fin) != NULL)
	{
		one_line[strlen(one_line) - 1] = '\0';
		if (!strcmp(one_line, "\0"))
			continue;

		if (!strcmp(one_line, "\t"))
			continue;

		if (!check_path(one_line))	{
			fprintf(stderr, "%s: %s\n", one_line, strerror(ENOENT));
			continue;
		}

		// 当路径为目录时
		if (path_is_dir(one_line))	{
			if (!recursive)	{		//非递归，跳过目录
				fprintf(stderr, "%s: %s\n", one_line, strerror(EISDIR));
				continue;	
			}
			else {		//递归，从目录继续获取文件列表
				kysec_get_file_list_from_path(list, one_line);
				continue;
			}
		}

		list_push_back(list, one_line);		//文件，直接添加至文件列表
	}
	fclose(fin);
	free(one_line);

	return 0;
}

static int kysec_set_label_for_path_list(P_Node **list, int label_type, char *new_label, int set)
{
	if (list == NULL)	{
		errno = EINVAL;
		return -1;
	}

	P_Node* cur = (*list); 
	while (cur != NULL) { 
		int ret = kysec_set_label_for_file(cur->fpath, label_type, new_label, set);
		if (ret != 0)
		{
			if (ret < 0)
				fprintf(stderr, "%s: %s\n", cur->fpath, strerror(errno));
			else
				fprintf(stderr, "%s[%d]: %s\n", cur->fpath, ret, get_dbus_err_msg());

			if (ret == 2 || ret == 1)
				break;
		}
		cur = cur->next; 
	}	

	return 0;
}

static int kysec_set_label_for_path(char *path, int label_type, char *new_label, int set)
{
	struct stat sb;

	if (!lstat(path, &sb) && S_ISDIR(sb.st_mode))
	{
		char *pathv[2] = {0};
		if (!recursive)	{
			errno = EISDIR;
			return -1;
		}

		pathv[0] = path;
		FTS *fts = fts_open(pathv, FTS_DEFAULT, NULL);
		FTSENT *ent = NULL;

		while (1)	{
			ent = fts_read(fts);
			if (!ent)	{
				if (errno != 0)
					continue;
				else
					break;
			}

			if (path_is_dir(ent->fts_path))	{
				//if (kysec_set_label_for_file(ent->fts_path, 'e', "trusted", set) < 0)		//目录不更改其trusted标记
				//fprintf(stderr, "%s: %s\n", ent->fts_path, strerror(errno));
				continue;
			}

			int ret = kysec_set_label_for_file(ent->fts_path, label_type, new_label, set);
			if (ret != 0)	{
				if (ret < 0)
					fprintf(stderr, "%s: %s\n", ent->fts_path, strerror(errno));
				else
					fprintf(stderr, "%s: %s\n", ent->fts_path, get_dbus_err_msg());
			}
			// if (kysec_set_label_for_file(ent->fts_path, label_type, new_label, set) < 0) {
			// 	fprintf(stderr, "%s: %s\n", ent->fts_path, strerror(errno));
			// 	continue;
			// }
		}

		fts_close(fts);
	}
	else
	{
		int ret = kysec_set_label_for_file(path, label_type, new_label, set);
		if (ret != 0)	{
			if (ret < 0)
				fprintf(stderr, "%s: %s\n", path, strerror(errno));
			else
				fprintf(stderr, "%s: %s\n", path, get_dbus_err_msg());
		}
		return ret;
	}
	//return kysec_set_label_for_file(path, label_type, new_label, set);
}

static int kysec_set_label_for_listfile(char *path, int label_type,
										char *new_label, int set)
{
	FILE *fin;
	char *one_line;

	one_line = (char *)malloc(PATH_MAX * sizeof(char));
	if ((fin = fopen(path, "r")) == NULL) {
		free(one_line);
		return -1;
	}

	while (fgets(one_line, PATH_MAX, fin) != NULL)
	{
		one_line[strlen(one_line) - 1] = '\0';
		if (!strcmp(one_line, "\0"))
			continue;
		if (!strcmp(one_line, "\t"))
			continue;
		if (!check_path(one_line))	{
			fprintf(stderr, "%s: %s\n", one_line, strerror(ENOENT));
			continue;
		}
		if (path_is_dir(one_line) && !recursive)
			fprintf(stderr, "%s: %s\n", one_line, strerror(EISDIR));

		if (kysec_set_label_for_path(one_line, label_type,
									 new_label, set) < 0)	{
			if (recursive)
				fprintf(stderr, "%s: %s\n", one_line,
						strerror(EPERM));
		}
	}
	fclose(fin);
	free(one_line);

	return 0;
}

static int kysec_set_label_for_package(char *pkg_name, int sid_type, char *new_label)
{
	char command[4097], linebuf[1024], *path = NULL;

	sprintf(command, "rpm -ql %s >/tmp/rpm-list 2>/dev/null", pkg_name);
	if (system(command))	{
		fprintf(stderr, "%s: package  not installed!\n", pkg_name);
		return -1;
	}

	FILE *fp = fopen("/tmp/rpm-list", "r");
	if (!fp)	{
		fprintf(stderr, "PACKAGE: package not installed!\n");
		return -1;
	}

	P_Node *list = NULL;
	while (fgets(linebuf, 1024, fp))	{
		linebuf[strlen(linebuf) - 1] = '\0';
		path = linebuf;
		struct stat st;
		stat(path, &st);

		if (S_ISDIR(st.st_mode) || !fopen(path, "r"))
			continue;

		list_push_back(&list, path);

		// if (kysec_set_label_for_path(path, sid_type, new_label, TRUE) < 0)
		// {
		// 	//fprintf(stderr, "%s: %s\n", path, strerror(errno));
		// 	return -1;
		// }
	}

	print_list(list);

	kysec_set_label_for_path_list(&list, sid_type, new_label, TRUE);
	destory_list(&list);

	fclose(fp);
	if (system("rm -f /tmp/rpm-list"))	{
		fprintf(stderr, "remove tmp file failed\n");
		return -1;
	}
	return 0;
}

int main(int argc, char **argv)
{
	// char cwd[1024] = {0};
	// getcwd(cwd, 1024);
	// printf("cwd = %s\n", cwd);

	int opt;
	char *path = NULL, *label_name, *label_val = NULL, *pkg_name = NULL;
	int option_index = 0;
	static struct option long_options[] = {
		{"package", required_argument, 0, 0},
		{0, 0, 0, 0}};

	while ((opt = getopt_long(argc, argv, "rxn:v:fh", long_options, &option_index)) != -1)
	{
		switch (opt)
		{
		case 0:
			pkg_name = optarg;
			break;
		case 'r':
			recursive = 1;
			break;
		case 'x':
			clear = 1;
			break;
		case 'n':
			label_name = optarg;
			break;
		case 'v':
			label_val = optarg;
			break;
		case 'f':
			listfile = 1;
			break;
		case 'h':
			usage();
			return 0;
		default:
			fprintf(stderr, "invalid arguments\n");
			usage();
			return -1;
		}
	}

	if (optind < 0 || !pkg_name && optind >= argc || !label_name)
		goto invalid;

	int sid_type = -1, ret = -1;
	kysec_sid_t new_sid;

#if 0
	if (!clear) {
		if (!label_val)
			goto invalid;

		if (!strcmp(label_name, "exectl"))
			sid_type = 'e';
		else if (!strcmp(label_name, "protect"))
			sid_type = 'p';
		else if (!strcmp(label_name, "userid"))
			sid_type = 'i';
		else
			fprintf(stderr, "Unknown label name: %s\n", label_name);

		if (pkg_name) {
			if (kysec_set_label_for_package(pkg_name, sid_type, label_val) < 0)
				return -1;
			return 0;
		}

		for (int i = optind; i < argc; i++) {
			path = argv[i];
			if (!listfile) {
				if (kysec_set_label_for_path(path, sid_type, label_val, TRUE) < 0) {
					//fprintf(stderr, "%s: %s\n", path, strerror(errno));
					if (optind == argc-1)
						return -1;
					continue;
				}
			} else {
				if (kysec_set_label_for_listfile(path, sid_type, label_val, TRUE) < 0) {
					fprintf(stderr, "%s: %s\n", path, strerror(errno));
					if (optind == argc - 1)
						return -1;
					continue;
				}
			}
		}
	} else {
		if (label_val)
			goto invalid;

		if (!strcmp(label_name, "exectl")) {
			sid_type = 'e';
			label_val = "unknown";
		} else if (!strcmp(label_name, "protect")) {
			sid_type = 'p';
			label_val = "none";
		} else if (!strcmp(label_name, "userid")) {
			sid_type = 'i';
			label_val = "none";
		} else {
			fprintf(stderr, "Unknown label name: %s\n", label_name);
			goto invalid;
		}

		for (int i = optind; i < argc; i++) {
			path = argv[i];
			if (!listfile) {
				if (kysec_set_label_for_path(path, sid_type, label_val, TRUE) < 0) {
					//fprintf(stderr, "%s: %s\n", path,
							// strerror(errno));
					if (optind == argc - 1)
						return -1;
					continue;
				}
			} else {
				if (kysec_set_label_for_listfile(path, sid_type, label_val, TRUE) < 0) {
					fprintf(stderr, "%s: %s\n", path,
							strerror(errno));
					if (optind == argc - 1)
						return -1;
					continue;
				}
			}
		}
	}
#endif

	if (!label_val && !clear)
		goto invalid;

	if (!strcmp(label_name, "exectl"))
		sid_type = 'e';
	else if (!strcmp(label_name, "protect"))
		sid_type = 'p';
	else if (!strcmp(label_name, "userid"))
		sid_type = 'i';
	else	{
		fprintf(stderr, "Unknown label name: %s\n", label_name);
		goto invalid;
	}

	if (clear)	{
		if (!strcmp(label_name, "exectl")) {
			label_val = "unknown";
		} else if (!strcmp(label_name, "protect") ||
			   !strcmp(label_name, "userid")) {
			label_val = "none";
		} else {
			fprintf(stderr, "Unknown label name: %s\n", label_name);
			goto invalid;
		}
	}

	if (pkg_name)	{
		if (kysec_set_label_for_package(pkg_name, sid_type, label_val) < 0)
			return -1;
		return 0;
	}

	P_Node* list = NULL;
	for (int i = optind; i < argc; i++) {
		path = argv[i];
		if (!listfile)	{
			if (kysec_get_file_list_from_path(&list, path) < 0)
				fprintf(stderr, "%s: %s\n", path, strerror(errno));
		}
		else {
			if (kysec_get_file_list_from_listfile(&list, path) < 0)
				fprintf(stderr, "%s: %s\n", path, strerror(errno));
		}

		// if (optind == argc - 1)
		// 	return -1;	
	}

	// print_list(list);

	kysec_set_label_for_path_list(&list, sid_type, label_val, TRUE);

	destory_list(&list);

	
	// for (int i = optind; i < argc; i++)
	// {
	// 	path = argv[i];
	// 	if (!listfile)
	// 	{
	// 		if (kysec_set_label_for_path(path, sid_type, label_val, TRUE) < 0)
	// 		{
	// 			//fprintf(stderr, "%s: %s\n", path, strerror(errno));
	// 			if (optind == argc - 1)
	// 				return -1;
	// 			continue;
	// 		}
	// 	}
	// 	else
	// 	{
	// 		if (kysec_set_label_for_listfile(path, sid_type, label_val, TRUE) < 0)
	// 		{
	// 			fprintf(stderr, "%s: %s\n", path, strerror(errno));
	// 			if (optind == argc - 1)
	// 				return -1;
	// 			continue;
	// 		}
	// 	}
	// }

	return 0;

invalid:
	usage();
	return -1;
}
