/*
 *  Kylin security execute control utils
 *
 *  Authors:  songchuanfei <songchuanfei@kylinos.cn>
 *  Authors:  huanfuxi <huanfuxi@kylinos.cn>
 *  Authors:  mengyuan <mengyuan@kylinos.cn>
 *
 *  Copyright (C) 2017,Tianjin KYLIN Information Technology Co., Ltd.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the
 *  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <kysec/status_dbus.h>
#include <kysec/libkysec_dbus.h>
#include <kysec/exectl.h>
#include <errno.h>

#define GRUB_PATH_CONFIG	"/etc/default/grub"
#define GRUB_PATH_CONFIG_BAK	"/etc/default/grub.bak"
#define SECURITY_CONFIG_ENV	"GRUB_CMDLINE_LINUX_SECURITY"

void usage(const char *progname)
{
	//fprintf(stderr, "usage:  %s < Softmode | Normal | Warning >\n", progname);
	fprintf(stderr, "usage:  %s < Softmode | Normal >\n", progname);
	fprintf(stderr, "\t%s < Softmode > [ -p ]\n", progname);
	fprintf(stderr, "\t%s -f  < exectl | fpro | kmod >  < on | off >\n", progname);
	exit(1);
}

static const char *func_name[] = {
	"exectl",
	"fpro",
	"kmod"
};

static int change_status_permanent(int status)
{
	int ret = 0;
	FILE *fp = NULL, *fp_bak = NULL;
	char *line = NULL;
	size_t len = 0;
	ssize_t nread;
	int find = 0;
	char security_str[512];

	fp = fopen(GRUB_PATH_CONFIG, "r");
	if (!fp) {
		fprintf(stderr, "Failed to open %s (%s)\n",
			GRUB_PATH_CONFIG, strerror(errno));
		return -1;
	}

	fp_bak = fopen(GRUB_PATH_CONFIG_BAK, "w+");
	if (!fp_bak) {
		fprintf(stderr, "Failed to open %s (%s)\n",
			GRUB_PATH_CONFIG_BAK, strerror(errno));
		fclose(fp);
		return -1;
	}

	while ((nread = getline(&line, &len, fp)) != -1) {
		if (line[0] == '#' || strstr(line, SECURITY_CONFIG_ENV) == NULL)
			fwrite(line, nread, 1, fp_bak);
		else {
			if (find == 0) {
				char *p = NULL, *q = NULL;
				memset(security_str, 0, sizeof(security_str));
				p = strstr(line, "kysec_status=");
				if (p) {
					q = p + strlen("kysec_status=") + 1;
					*p = '\0';
					snprintf(security_str, sizeof(security_str), "%skysec_status=%d%s",
						 line, status, q);
				} else {
					if (!strstr(line, "security=")) {
						/* GRUB_CMDLINE_LINUX_SECURITY= or GRUB_CMDLINE_LINUX_SECURITY="xxxx" */
						p = strstr(line, SECURITY_CONFIG_ENV);
						q = p + strlen(SECURITY_CONFIG_ENV);
						*p = '\0';
						if (*(q + 1) == '"')
							snprintf(security_str, sizeof(security_str), "%s=\"kysec_status=%d %s",
								 SECURITY_CONFIG_ENV, status, q + 2);
						else
							snprintf(security_str, sizeof(security_str), "%s=\"kysec_status=%d\"\n",
								 SECURITY_CONFIG_ENV, status);
					} else {
						/* GRUB_CMDLINE_LINUX_SECURITY="security=xxx" */
						p = strrchr(line, '"');
						*p = '\0';
						snprintf(security_str, sizeof(security_str), "%s kysec_status=%d\"\n",
							 line, status);
					}
				}
				//printf("%s", security_str);
				fwrite(security_str, strlen(security_str), 1, fp_bak);
				find++;
			} else
				continue;
		}
	}

	free(line);
	fclose(fp);
	fclose(fp_bak);

	ret = rename(GRUB_PATH_CONFIG_BAK, GRUB_PATH_CONFIG);
	if (ret == -1) {
		printf("Failed to rename: %s\n", strerror(errno));
		remove(GRUB_PATH_CONFIG_BAK);
		return -1;
	}

	return 0;
}

int main(int argc, char **argv)
{
	int rc = 0, opt, status, son = 0;
	char *kysec_name;
	int permanent = 0;
	int kysec_status = KYSEC_STATUS_SOFTMODE;
	char *cmd = NULL;

	while ((opt = getopt(argc, argv, "f:p")) != -1) {
		switch (opt) {
		case 'f':
			son = 1;
			kysec_name = optarg;
			break;
		case 'p':
			permanent = 1;
			break;
		default:
			usage(argv[0]);
			return -1;
		}
	}

	if (son) {
		int func_id;

		if (!argv[3]) {
			usage(argv[0]);
			return -1;
		}

		for (func_id = 0; func_id < sizeof(func_name)/sizeof(func_name[0]); func_id++) {
			if (!strcmp(func_name[func_id], kysec_name))
				break;
		}

		if (!strcmp(argv[3], "on")) {
			status = FUNC_ENABLED;
		} else if (!strcmp(argv[3], "off")) {
			status = FUNC_DISABLED;
		} else {
			usage(argv[0]);
			return -1;
		}

		if (kysec_set_func_status(func_id, status) < 0) {
			fprintf(stderr, "%s: set %s status failed\n",
				"kysec_set", kysec_name);
			return 2;
		}
		return 0;
	} else {
		if (argc != 2 && argc != 3) {
			usage(argv[0]);
			return -1;
		}

		if (permanent && kysec_get_3adm_status()) {
			fprintf(stderr, "WARNNING: '-p' is an invalid argument, when 3adm is enabled\n");
			return -1;
		}

		if (argc == 2)
			cmd = argv[1];
		else if (argc == 3)
			cmd = argv[2];

		if (strcasecmp(cmd, "normal") == 0) {
			if (permanent) {
				fprintf(stderr, "WARNNING: the '-p' argument is an invalid argument, when there is a 'normal' argument\n");
				return -1;
			}
			rc = kysec_setstatus(KYSEC_STATUS_NORMAL);
			kysec_status = KYSEC_STATUS_NORMAL;
		} else if (strcasecmp(cmd, "softmode") == 0) {
			rc = kysec_setstatus(KYSEC_STATUS_SOFTMODE);
			kysec_status = KYSEC_STATUS_SOFTMODE;
		}/* else if (strcasecmp(cmd, "warning") == 0) {
			rc = kysec_setstatus(KYSEC_STATUS_WARNING);
			kysec_status = KYSEC_STATUS_WARNING;
		} */else {
			usage(argv[0]);
			return -1;
		}

		if (rc != 0) {
			fprintf(stderr, "%s: %s\n",	argv[0], get_dbus_err_msg());
			return 1;
		}

		// if (rc == -2) {
		// 	fprintf(stderr, "%s: Kysec is disabled\n",
		// 		argv[0]);
		// 	return 2;
		// } else if (rc == -1) {
		// 	fprintf(stderr, "%s: %s\n",
		// 		argv[0], strerror(errno));
		// 	return 1;
		// }

		rc = 0;
		if (permanent) {
			if (geteuid() != USER_SYSADM) {
				errno = ENOTSUP;
				fprintf(stderr, "%s: %s\n", argv[0], strerror(errno));
				return -1;
			}

			rc = change_status_permanent(kysec_status);
			if (rc)
				return rc;

			/* update-grub */
			rc = system("update-grub");
		}

		return rc;
	}
}
